## ------ language="Python" file="entangled/config.py"
from .typing import (JSONType)
import subprocess
import json
import sys

def read_config() -> JSONType:
    """Reads config from `entangled.dhall` with fall-back to `entangled.json`."""
    try:
        result = subprocess.run(
            ["dhall-to-json", "--file", "entangled.dhall"],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8', check=True)
        return json.loads(result.stdout)
    except subprocess.CalledProcessError as e:
        print("Error reading `entangled.dhall`:\n" + e.stderr, file=sys.stderr)
    except FileNotFoundError:
        print("Warning: could not find `dhall-to-json`, trying to read JSON instead.",
              file=sys.stderr)
    return json.load(open("entangled.json", "r"))

def get_language_info(config: JSONType, identifier: str) -> JSONType:
    try:
        return next(lang for lang in config["languages"]
                    if identifier in lang["identifiers"])
    except StopIteration:
        raise ValueError(f"Language with identifier `{identifier}` not found in config.")
## ------ end



## ------ language="Python" file="entangled/annotate.py"
from collections import defaultdict
from .tangle import get_name
from panflute import (Span, Str, Para, CodeBlock, Div, Emph)

def prepare(doc):
    doc.code_count = defaultdict(lambda: 0)

def action(elem, doc):
    if isinstance(elem, CodeBlock):
        name = get_name(elem)
        if name is None:
            return
        if doc.code_count[name] == 0:
            label = Span(Emph(Str(f"«{name}»=")))
            doc.code_count[name] += 1
        else:
            label = Span(Emph(Str(f"«{name}»+")))
        return Div(Para(label), elem, classes=["annotated-code"])
## ------ end



## ------ language="Python" file="entangled/doctest.py"
from panflute import (Doc, Element, CodeBlock)
from .typing import (ActionReturn, JSONType, CodeMap)
from .tangle import (get_name, expand_code_block)
from .config import get_language_info
from collections import defaultdict

import sys

## ------ begin <<doctest-suite>>[0]
from dataclasses import dataclass
from typing import (Optional, List, Dict)
from enum import Enum

class TestStatus(Enum):
    PENDING = 0
    SUCCESS = 1
    FAIL = 2
    ERROR = 3
    UNKNOWN = 4
## ------ end
## ------ begin <<doctest-suite>>[1]
@dataclass
class Test:
    __test__ = False    # not a pytest class
    code: str
    expect: Optional[str]
    result: Optional[str] = None
    error: Optional[str] = None
    status: TestStatus = TestStatus.PENDING
## ------ end
## ------ begin <<doctest-suite>>[2]
@dataclass
class Suite:
    code_blocks: List[Test]
    language: str
## ------ end
## ------ begin <<get-doc-tests>>[0]
def get_language(c: CodeBlock) -> str:
    if not c.classes:
        raise ValueError(f"Code block `{c.name}` has no language specified.")
    return c.classes[0]

def get_doc_tests(code_map: CodeMap) -> Dict[str, Suite]:
    def convert_code_block(c: CodeBlock) -> Test:
        name = get_name(c)
        code = expand_code_block(code_map, c)
        if "doctest" in c.classes:
            s = code.split("\n---\n")
            if len(s) != 2:
                raise ValueError(f"Doc test `{name}` should have single `---` line.")
            return Test(*s)
        else:
            return Test(code, None)

    result = {}
    for k, v in code_map.items():
        if any("doctest" in c.classes for c in v):
            result[k] = Suite(
                code_blocks=[convert_code_block(c) for c in v],
                language=get_language(v[0]))

    return result
## ------ end
## ------ begin <<doctest-report>>[0]
from panflute import Div

def generate_report(elem: CodeBlock, t: Test) -> ActionReturn:
    lang_class = elem.classes[0]

    ## ------ begin <<doctest-content-div>>[0]
    def content_div(*output):
        status_attr = {"status": t.status.name}
        input_code = Div(CodeBlock(
            t.code, identifier=elem.identifier,
            classes=elem.classes), classes=["doctestInput"])
        return Div(input_code, *output, classes=["doctest"], attributes=status_attr)
    ## ------ end
    if t.status is TestStatus.ERROR:
        return content_div( Div( CodeBlock(str(t.error))
                               , classes=["doctestError"] ) )
    if t.status is TestStatus.FAIL:
        return content_div( Div( CodeBlock(str(t.result), classes=[lang_class])
                               , classes=["doctestResult"] )
                          , Div( CodeBlock(str(t.expect), classes=[lang_class])
                               , classes=["doctestExpect"] ) )
    if t.status is TestStatus.SUCCESS:
        return content_div( Div( CodeBlock(str(t.result), classes=[lang_class])
                               , classes=["doctestResult"] ) )
    if t.status is TestStatus.PENDING:
        return content_div()
    if t.status is TestStatus.UNKNOWN:
        return content_div( Div( CodeBlock(str(t.result))
                               , classes=["doctestUnknown"] ) )
    return None
## ------ end
## ------ begin <<doctest-run-suite>>[0]
import jupyter_client
import queue

def run_suite(config: JSONType, s: Suite) -> None:
    ## ------ begin <<jupyter-get-kernel-name>>[0]
    info = get_language_info(config, s.language)
    kernel_name = info["jupyter"] if "jupyter" in info else None
    if not kernel_name:
        raise RuntimeError(f"No Jupyter kernel known for the {s.language} language.")
    specs = jupyter_client.kernelspec.find_kernel_specs()
    if kernel_name not in specs:
        raise RuntimeError(f"Jupyter kernel `{kernel_name}` not installed.")
    ## ------ end
    with jupyter_client.run_kernel(kernel_name=kernel_name) as kc:
        print(f"Kernel `{kernel_name}` running ...", file=sys.stderr)
        ## ------ begin <<jupyter-eval-test>>[0]
        def jupyter_eval(test: Test):
            msg_id = kc.execute(test.code)
            while True:
                try:
                    msg = kc.get_iopub_msg(timeout=1000)
                    if handle(test, msg_id, msg):
                       return
        
                except queue.Empty:
                    test.error = "Operation timed out."
                    test.status = TestStatus.ERROR
                    return
        ## ------ end
        ## ------ begin <<jupyter-eval-test>>[1]
        def handle(test, msg_id, msg):
            from pampy import match, _
            ## ------ begin <<jupyter-handlers>>[0]
            def execute_result_text(data):
                test.result = data
                if (test.expect is None) or test.result.strip() == test.expect.strip():
                    test.status = TestStatus.SUCCESS
                else:
                    test.status = TestStatus.FAIL
                return True
            ## ------ end
            ## ------ begin <<jupyter-handlers>>[1]
            def status_idle(_):
                if test.expect is None:
                    test.status = TestStatus.SUCCESS
                else:
                    test.status = TestStatus.FAIL
                return True
            ## ------ end
            ## ------ begin <<jupyter-handlers>>[2]
            def error_traceback(tb):
                test.error = "\n".join(msg["content"]["traceback"])
                test.status = TestStatus.ERROR 
                return True
            ## ------ end
            return match(msg
                ## ------ begin <<jupyter-match>>[0]
                , { "msg_type": "execute_result"
                  , "parent_header": { "msg_id" : msg_id }
                  , "content": { "data" : { "text/plain": _ } } }
                , execute_result_text
                ## ------ end
                ## ------ begin <<jupyter-match>>[1]
                , { "msg_type": "status"
                  , "parent_header": { "msg_id" : msg_id }
                  , "content": { "execution_state": "idle" } }
                , status_idle
                ## ------ end
                ## ------ begin <<jupyter-match>>[2]
                , { "msg_type": "error"
                  , "parent_header": { "msg_id" : msg_id }
                  , "content": { "traceback": _ } }
                , error_traceback
                ## ------ end
                ## ------ begin <<jupyter-match>>[3]
                , _
                , lambda x: False
                ## ------ end
                )
        ## ------ end

        for test in s.code_blocks:
            jupyter_eval(test)
            if test.status is TestStatus.ERROR:
                break
## ------ end

def prepare(doc: Doc) -> None:
    assert hasattr(doc, "config"), "Need to read config first."
    assert hasattr(doc, "code_map"), "Need to tangle first."
    doc.suites = get_doc_tests(doc.code_map)
    for name, suite in doc.suites.items():
        run_suite(doc.config, suite)
    doc.code_counter = defaultdict(lambda: 0)

def action(elem: Element, doc: Doc) -> ActionReturn:
    if isinstance(elem, CodeBlock):
        name = get_name(elem)
        if "doctest" in elem.classes:
            suite = doc.suites[name].code_blocks[doc.code_counter[name]]
            doc.code_counter[name] += 1
            return generate_report(elem, suite)
        doc.code_counter[name] += 1
    return None
## ------ end



## ------ language="Python" file="entangled/typing.py"
from typing import (Union, List, Dict, Callable, Any)
from panflute import (Element, Doc, CodeBlock)

ActionReturn = Union[Element, List[Element], None]
Action = Callable[[Element, Doc], ActionReturn]
CodeMap = Dict[str, List[CodeBlock]]
JSONType = Any
## ------ end



## ------ language="Python" file="entangled/doctest_main.py"
import panflute

from . import tangle
from . import annotate
from . import doctest

from .config import read_config


def main() -> None:
    ## ------ begin <<load-document>>[0]
    import io
    import sys
    
    json_input = sys.stdin.read()
    json_stream = io.StringIO(json_input)
    doc = panflute.load(json_stream)
    ## ------ end
    doc.config = read_config()

    tangle.prepare(doc)
    doc = doc.walk(tangle.action)

    annotate.prepare(doc)
    doc = doc.walk(annotate.action)

    doctest.prepare(doc)
    doc = doc.walk(doctest.action)

    panflute.dump(doc)
## ------ end



## ------ language="Python" file="entangled/tangle.py"
from panflute import (run_filter, Doc, Element, CodeBlock)
from typing import (Optional, Dict, Callable)
from .typing import (CodeMap)
import sys

## ------ begin <<get-code-block>>[0]
import re

## ------ begin <<replace-expr>>[0]
def replace_expr(expr: str, replace: Callable[..., str], text: str) -> str:
    """Matches (fullmatch) `text` using the expression `expr`. If the expression
    matches, then returns the result of passing named sub-matches as
    keyword arguments to `replace`. Returns `text` otherwise."""
    match = re.fullmatch(expr, text)
    if match:
        return replace(**match.groupdict())
    else:
        return text
## ------ end

def get_code(code_map: CodeMap, name: str) -> str:
    ## ------ begin <<expand>>[0]
    def expand(code: CodeBlock) -> str:
        pattern = "(?P<prefix>[ \t]*)<<(?P<name>[^ >]*)>>\\Z"
        return "\n".join(
            replace_expr(pattern, look_up, line)
            for line in code.text.splitlines())
    ## ------ end
    ## ------ begin <<look-up>>[0]
    from textwrap import indent
    
    def look_up(*, name: str, prefix: str) -> str:
        blocks = code_map[name]
        if not blocks:
            raise ValueError(f"No code with name `{name}` found.")
        result = "\n".join(expand(code) for code in blocks)
        return indent(result, prefix)
    ## ------ end
    return look_up(name=name, prefix="")

def expand_code_block(code_map: CodeMap, code_block: CodeBlock) -> str:
    ## ------ begin <<expand>>[0]
    def expand(code: CodeBlock) -> str:
        pattern = "(?P<prefix>[ \t]*)<<(?P<name>[^ >]*)>>\\Z"
        return "\n".join(
            replace_expr(pattern, look_up, line)
            for line in code.text.splitlines())
    ## ------ end
    ## ------ begin <<look-up>>[0]
    from textwrap import indent
    
    def look_up(*, name: str, prefix: str) -> str:
        blocks = code_map[name]
        if not blocks:
            raise ValueError(f"No code with name `{name}` found.")
        result = "\n".join(expand(code) for code in blocks)
        return indent(result, prefix)
    ## ------ end
    return expand(code_block)
## ------ end

## ------ begin <<tangle-prepare>>[0]
from collections import defaultdict

def prepare(doc: Doc) -> None:
    doc.code_map = defaultdict(list)
## ------ end
## ------ begin <<tangle-action>>[0]
## ------ begin <<get-name>>[0]
def get_name(elem: Element) -> Optional[str]:
    if elem.identifier:
        return elem.identifier

    if "file" in elem.attributes:
        return elem.attributes["file"]

    return None
## ------ end

def action(elem: Element, doc: Doc) -> None:
    if isinstance(elem, CodeBlock):
        name = get_name(elem)
        if name:
            doc.code_map[name].append(elem)
## ------ end
## ------ begin <<tangle-finalize>>[0]
def get_file_map(code_map: CodeMap) -> Dict[str, str]:
    """Extracts all file references from `code_map`."""
    return { code[0].attributes["file"]: codename 
             for codename, code in code_map.items()
             if "file" in code[0].attributes }
## ------ end
## ------ begin <<tangle-finalize>>[1]
def write_file(filename: str, text: str) -> None:
    """Writes `text` to file `filename`, only if `text` is different
    from contents of `filename`."""
    try:
        content = open(filename).read()
        if content == text:
            return
    except FileNotFoundError:
        pass
    print(f"Writing `{filename}`.", file=sys.stderr)
    open(filename, 'w').write(text)

def finalize(doc: Doc) -> None:
    """Writes all file references found in `doc.code_map` to disk.
    This only overwrites a file if the content is different."""
    file_map = get_file_map(doc.code_map)
    for filename, codename in file_map.items():
        write_file(filename, get_code(doc.code_map, codename))
    doc.content = []
## ------ end

def main(doc: Optional[Doc] = None) -> None:
    run_filter(
        action, prepare=prepare, finalize=finalize, doc=doc)
## ------ end



#!/usr/bin/env python

from setuptools import setup, find_packages

test_deps = [
      "pytest>=5,<6"
    , "pytest-cov>=2.8.1,<3"
    , "pytest-mypy>=0.4.2,<1"
    , "flake8>=3,<4"
    , "ipykernel>=5.1.3,<6" ]

setup(
    name="entangled-filters",
    version="0.4",
    packages=find_packages(),

    install_requires=[
          "jupyter_client>=5.3.4,<6"
        , "pampy>=0.3.0,<0.4"
        , "panflute>=1.11.2,<1.12"
    ],
    tests_require=test_deps,
    extras_require={
        "test": test_deps
    },
    # include_package_data=True,
    # package_data={
    #     # If any package contains *.txt or *.rst files, include them:
    #     '': ['*.txt', '*.rst'],
    #     # And include any *.msg files found in the 'hello' package, too:
    #     'hello': ['*.msg'],
    # },

    # metadata to display on PyPI
    author="Johan Hidding",
    author_email="j.hidding@esciencecenter.nl",
    description="Set of Pandoc filters to aid in literate programming",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    keywords="literate programming",
    url="https://entangled.github.io/",   # project home page, if any

    project_urls={
          "Bug Tracker": "https://github.com/entangled/filters/issues/"
        , "Documentation": "https://entangled.github.io/filters/"
        , "Source Code": "https://github.com/entangled/filters/"
    },

    classifiers=[
          'License :: OSI Approved :: Apache Software License'
        , 'Development Status :: 3 - Alpha'
        , 'Intended Audience :: Science/Research'
        , 'Intended Audience :: Developers'
        , 'Environment :: Console'
        , 'Natural Language :: English'
        , 'Operating System :: OS Independent'
        , 'Programming Language :: Python :: 3.7'
        , 'Programming Language :: Python :: 3.8'
        , 'Topic :: Education'
        , 'Topic :: Documentation'
        , 'Topic :: Scientific/Engineering'
        , 'Topic :: Software Development'
        , 'Topic :: Software Development :: Documentation'
        , 'Topic :: Software Development :: Testing'
        , 'Topic :: Text Processing :: Markup'
    ],

    entry_points = {
        'console_scripts': [
              'pandoc-tangle=entangled.tangle:main'
            , 'pandoc-test=entangled.doctest_main:main'
        ]
    }
    # could also include long_description, download_url, etc.
)

